/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "segment.hh"
#include "buffer.hh"
#include "test_endian.hh"
#include "gds_atomic.h"
#include <stdexcept>
#include <iostream>
#include <cstring>

using namespace sends;
using namespace std;

static test_endian bOrder;

//======================================  Segment default constructor
segment::segment(void) 
    : mBufClass(-1), mData(0)
{
    clear();
}

//======================================  Segment destructor
segment::~segment(void) {
    release();
}

//======================================  Reset the segment maximum time
void
segment::cap(gps_type tMax) {
    if (tMax > mTMax || tMax < mStop || tMax < mReserved) {
	cerr << "Segment cap error, tMax: " << tMax << endl;
	dump_header(cerr);
	dump(cerr);
	throw invalid_argument("Invalid cap time");
    } else {
	cout << "Segment for chan: " << mChannel << " times " << mStart 
	     << ":" << mTMax << " capped at: " << tMax << endl;
    }
    mTMax = tMax;
}

//======================================  Clear the segment (retain the buffer)
void
segment::clear(void) {
    mChannel  =  ~0;
    mStart    =   0;
    mStop     =   0;
    mReserved =   0;
    mTMax     =   0;
    mRate     = 0.0;
    mWordSize =   0;
    mUseCount =   0;
    mBias     = 0.0;
    mScale    = 1.0;
    mUnits.clear();
}

//======================================  Calculate data position
index_type
segment::dataIndex(gps_type gps) const {
    double rw = mRate;
    if (gps < mStart || rw <= 0 || mWordSize <= 0) {
	cerr << "Start: " << mStart << " t: "    << gps 
	     << " rate: " << mRate  << " size: " << mWordSize << endl;
	throw runtime_error("Segment info not valid");
    }
    return static_cast<index_type>(rw * (gps - mStart) + 0.5) * mWordSize;
}

//======================================  Dump a segment
std::ostream&
segment::dump(std::ostream& out) const {
    out << mChannel  << "  " << mBufClass << "  " << mStart << "  " 
	<< mStop     << "  " << mReserved << "  " << mTMax  << "  " 
	<< mRate     << "  " << mWordSize << "  " << mUseCount << endl;
    return out;
}

//======================================  Print a segment dump header
std::ostream&
segment::dump_header(std::ostream& out) {
    out << "chan  bClass  start  stop reserved max rate size use" << endl;
    return out;
}

//======================================  Extend the segment data field.
void 
segment::extend(const void* data, count_type inx, count_type length) {
    char* p = refData(mStop);
    count_type lBytes = length * mWordSize;
    mData->push(lBytes);
    const char* in = reinterpret_cast<const char*>(data) + inx * mWordSize;

    //-----------------------------------  memcpy if big-endian or byte data
    if (bOrder.big_end() || mWordSize < 2) {
	memcpy(p, in, lBytes);

    //-----------------------------------  Swap bytes.
    } else {
	for (count_type i=0; i<length; ++i) {
	    for (count_type j=mWordSize; j--;) {
		*p++ = in[j];
	    }
	    in += mWordSize;
	}
    }
    int dt = int(length / mRate + 0.5);
    mStop += dt;
}

//======================================  Lock/Ulock the fill mutex
void
segment::fill_lock(void) {
    mFillMutex.writelock();
}

void
segment::fill_unlock(void) {
    mReserved = mStop;
    mFillMutex.unlock();
}

//======================================  Get a pointer to the data
const char* 
segment::refData(gps_type gps) const {
    return mData->data() + dataIndex(gps);
}

char* 
segment::refData(gps_type gps) {
    return mData->data() + dataIndex(gps);
}

//======================================  Clear the segment
void
segment::release(void) {
    if (mData) {
	mData->free();
	mData =   0;
    }
    mBufClass =  -1;
    clear();
}

//======================================  Reserve space in the data field.
bool
segment::reserve(gps_type start, gps_type stop) {
    if (start < mStart || start >= mTMax) return false;
    fill_lock();
    if (start > mStop) {
	fill_unlock();
	return false;
    }
    if (stop > mTMax) mReserved = mTMax;
    else              mReserved = stop;
    return true;
}

//======================================  Get a buffer
void 
segment::set_buffer(buffer* bufpt, int bufclass) {
    release();
    mBufClass = bufclass;
    mData     = bufpt;
}

//======================================  Initialize the segment
void 
segment::set_segment(chan_index chan, double rate, int dLen, gps_type start) {
    if (!mData) throw runtime_error("Segment buffer not allocated");
    mChannel  = chan;
    mRate     = rate;
    mWordSize = dLen;
    mStart    = start;
    mStop     = start;
    mReserved = start;
    long nSec = long((mData->capacity()/mWordSize)/mRate);
    if (mRate < 1.0) {
	int step = int(1.0/mRate + 0.5);
	nSec -= nSec%step;
    }
    mTMax = mStart + nSec;
}

//======================================  Initialize the segment
void 
segment::set_units(double bias, double scale, const std::string& units) {
    mBias  = bias;
    mScale = scale;
    mUnits = units;
}

//======================================  Increment the use count
segment::count_type
segment::unuse(void) {
    if (!mUseCount) throw logic_error("Decrement use count of unused segment");
#ifdef USE_BUILTIN_ATOMICS
    return ATOMIC_PREDEC(mUseCount);
#else
    thread::semlock lock(mUseMutex);
    return --mUseCount;
#endif
}

//======================================  Increment the use count
void 
segment::use(void) {
#ifdef USE_BUILTIN_ATOMICS
    ATOMIC_POSTINC(mUseCount);
#else
    thread::semlock lock(mUseMutex);
    mUseCount++;
#endif
}
